-- by mor2000

--------------------
-- Mover (changes translation, rotation and visibility of objects)

UPK_Mover={}
local UPK_Mover_mt = Class(UPK_Mover, UniversalProcessKit)
InitObjectClass(UPK_Mover, "UPK_Mover")
UniversalProcessKit.addModule("mover",UPK_Mover)

function UPK_Mover:new(isServer, isClient, customMt)
	if customMt == nil then
		customMt = UPK_Mover_mt
	end
	local self = UniversalProcessKit:new(isServer, isClient, customMt)
	registerObjectClassName(self, "UPK_Mover")
	return self
end

function UPK_Mover:load(id,parent)
	UPK_FillTrigger:superClass().load(self, id, parent)
	
    local function tobool(val)
  		return not (val == nil or val == false or val == 0 or val == "0" or val == "false" )
    end
	
	local function getVectorFromUserAttribute(attribute, default)
		local str=Utils.getNoNil(getUserAttribute(id, attribute), default)
		if type(str)=="string" then
			return __c({Utils.getVectorFromString(str)})
		end
		return str
	end
	
	local function gmatch(str, pattern)
		local arr={}
		for v in string.gmatch(str,pattern) do
			table.insert(arr,v)
		end
		return arr
	end
	
    local fillTypeString = Utils.getNoNil(getUserAttribute(id, "fillTypes"), "unknown")
	self.fillTypes = UniversalProcessKit.fillTypeNameToInt(gmatch(fillTypeString,"%S+"))
	self.fillTypeChoiceMax = Utils.getNoNil(getUserAttribute(id, "fillTypeChoice"), "max")=="max"
	
	-- move
	self.startMovingAt=Utils.getNoNil(tonumber(getUserAttribute(id, "startMovingAt")), 0)
	self.stopMovingAt=Utils.getNoNil(tonumber(getUserAttribute(id, "stopMovingAt")), self.capacity)
	local posMin = getVectorFromUserAttribute("lowPosition", "0 0 0")
	local posMax = getVectorFromUserAttribute("highPosition", posMin)
	local posLower = getVectorFromUserAttribute("lowerPosition", posMin)
	local posHigher = getVectorFromUserAttribute("higherPosition", posMax)
	self.posMin = self.pos + posMin
	self.posMax = self.pos + posMax
	self.posLower = self.pos + posLower
	self.posHigher = self.pos + posHigher
	self.moveType = Utils.getNoNil(getUserAttribute(id, "moveType"), "linear")
	
	-- rotation
	self.startRotatingAt=Utils.getNoNil(tonumber(getUserAttribute(id, "startRotatingAt")), 0)
	self.stopRotatingAt=Utils.getNoNil(tonumber(getUserAttribute(id, "stopRotatingAt")), self.capacity)
	local rpsMin = getVectorFromUserAttribute("lowRotationsPerSecond", "0 0 0")
	local rpsMax = getVectorFromUserAttribute("highRotationsPerSecond", rpsMin)
	local rpsLower = getVectorFromUserAttribute("lowerRotationsPerSecond", rpsMin)
	local rpsHigher = getVectorFromUserAttribute("higherRotationsPerSecond", rpsMax)
	self.rpsMin = rpsMin*(2*math.pi)
	self.rpsMax = rpsMax*(2*math.pi)
	self.rpsLower = rpsLower*(2*math.pi)
	self.rpsHigher = rpsHigher*(2*math.pi)
	self.rotationType = Utils.getNoNil(getUserAttribute(id, "rotationType"), "linear")
	
	-- visibility
	self.startVisibilityAt=Utils.getNoNil(tonumber(getUserAttribute(id, "startVisibilityAt")), 0)
	self.stopVisibilityAt=Utils.getNoNil(tonumber(getUserAttribute(id, "stopVisibilityAt")), self.capacity)
	self.visibilityType=Utils.getNoNil(getUserAttribute(id, "visibilityType"), "show")=="show"
	
    return true
end

function UPK_Mover:delete()
	UPK_Mover:superClass().delete(self)
end

function UPK_Mover:update(dt)
	UPK_Mover:superClass().update(self,dt)
	
	local newFillLevel
	if self.fillTypeChoiceMax then
		newFillLevel= self.fillLevels(self.fillTypes):max() or 0
	else
		newFillLevel= self.fillLevels(self.fillTypes):min() or 0
	end
	newFillLevel=math.min(math.max(newFillLevel,0),self.capacity)

	-- move only if sth changed
	if newFillLevel~=self.oldFillLevel then
		if newFillLevel<self.startMovingAt then
			self.pos=self.posLower
		elseif newFillLevel>self.stopMovingAt then
			self.pos=self.posHigher
		else
			local ratio=self:getRatio("pos",self.moveType,newFillLevel,self.startMovingAt,self.stopMovingAt)
			self.pos=self.posMin+(self.posMax-self.posMin)*ratio
		end
		setTranslation(self.nodeId,unpack(self.pos))
		
		-- rotation
		if newFillLevel<self.startRotatingAt then
			self.rotStep=self.rpsLower
		elseif newFillLevel>self.stopRotatingAt then
			self.rotStep=self.rpsHigher
		else
			local rotRatio=self:getRatio("rot",self.rotationType,newFillLevel,self.startRotatingAt,self.stopRotatingAt)
			self.rotStep=self.rpsMin+(self.rpsMax-self.rpsMin)*rotRatio
		end
		
		-- visibility
		setVisibility(self.nodeId,self.visibilityType==(newFillLevel>=self.startVisibilityAt and newFillLevel<=self.stopVisibilityAt))
	end

	-- rotate all the time
	rotate(self.nodeId, unpack(self.rotStep*(dt*0.001)))

	self.oldFillLevel=newFillLevel
end

function UPK_Mover:getRatio(use,type,fillLevel,minFillLevel,maxFillLevel)
	if minFillLevel<0 or maxFillLevel<0 then
		return 0
	end
	local dividend
	if self.ratioMaxFillLevel==nil then
		self.ratioMaxFillLevel={}
	end
	if self.ratioMaxFillLevel[use]== nil then
		self.ratioMaxFillLevel[use]={}
		self.ratioMaxFillLevel[use].sphere=((maxFillLevel-minFillLevel)/(4/3*math.pi))^(1/3)
		self.ratioMaxFillLevel[use].cone=((maxFillLevel-minFillLevel)/(1/3*math.pi))^(1/3)
		self.ratioMaxFillLevel[use].square=(maxFillLevel-minFillLevel)^(1/2)
		self.ratioMaxFillLevel[use].circle=((maxFillLevel-minFillLevel)/math.pi)^(1/2)
		self.ratioMaxFillLevel[use].sinus=1
		self.ratioMaxFillLevel[use].linear=maxFillLevel-minFillLevel
	end
	if type=="sphere" then
		dividend=((fillLevel-minFillLevel)/(4/3*math.pi))^(1/3)
	elseif type=="cone" then
		dividend=((fillLevel-minFillLevel)/(1/3*math.pi))^(1/3)
	elseif type=="square" then
		dividend=(fillLevel-minFillLevel)^(1/2)
	elseif type=="circle" then
		dividend=((fillLevel-minFillLevel)/math.pi)^(1/2)
	elseif type=="sinus" then
		dividend=math.sin((fillLevel-minFillLevel)/(maxFillLevel-minFillLevel)*math.pi)
	else
		type="linear"
		dividend=fillLevel-minFillLevel
	end
	return dividend/self.ratioMaxFillLevel[use][type]
end

